import java.util.*;

public class Cache extends Hashtable {

    private int maxHits;
    private long maxTime;
    private boolean activeCache = true;

    public Cache (int mh, long mt) {
        maxHits = mh;
        if (mt == -1) maxTime = Long.MAX_VALUE;
        else maxTime = mt;
    }

    public Cache (long mt) {
        this(Integer.MAX_VALUE,mt);
    }

    public Cache () {
        this(Integer.MAX_VALUE,60000);
    }

    public void setMaxHits (int mh) {
        if (mh == -1) maxHits = Integer.MAX_VALUE;
        else maxHits = mh;
    }

    public void setMaxTime (long mt) {
        if (mt == -1) maxTime = Long.MAX_VALUE;
        else maxTime = mt*1000;
    }
    
    public int getMaxHits () {
        return maxHits == Integer.MAX_VALUE ? -1 : maxHits;
    }

    public long getMaxTime () {
        return maxTime == Long.MAX_VALUE ? -1 : (maxTime/1000);
    }


    public void setCacheInactive () {
        activeCache = false;
        expireCache();
    }
    
    public void setCacheActive () {
        activeCache = true;
    }
    
    public boolean isCacheActive () {
        return (activeCache);
    }

/*    public int getCacheSize () {
        Enumeration en = super.keys();
        String tmpS;
        int cs = 0;
        while (en.hasMoreElements()) {
            ExceptionHandler.debugInfo("1");
            tmpS = (String)super.get(en.nextElement());
            ExceptionHandler.debugInfo("2");
            if (tmpS!=null) cs+=tmpS.length();
            ExceptionHandler.debugInfo("3");
        }
        return cs;
    }
*/   
    public synchronized boolean containsKey(Object key) {
        if (super.containsKey(key)) {
            CacheElement ce = (CacheElement)super.get(key);  
            return ((ce.numberOfHits < maxHits) && (ce.getLiveTime() < maxTime));
        }
        return false;
           
    }


    private void removeExpiredElements(){
        Enumeration en = keys();
        while (en.hasMoreElements()) {
            Object key = en.nextElement();
            CacheElement ce = (CacheElement)super.get(key);  
            if ((ce.numberOfHits > maxHits) || (ce.getLiveTime() > maxTime))            
               expireElement(key);
        }

    }

    public synchronized Object get (Object key) {
        removeExpiredElements();
        if (super.containsKey(key)) {
            CacheElement ce = (CacheElement)super.get(key);  
            if (!((ce.numberOfHits > maxHits) || (ce.getLiveTime() > maxTime))) {
                 ce.numberOfHits++;
                 super.put(key,ce);
                 return ce.contents;  
            }
        }
        return null;
    }

	public synchronized Object put(Object key,Object value) throws NullPointerException {
        if (!activeCache) return(null);
        CacheElement ce = new CacheElement(value, 0, System.currentTimeMillis());
        return (super.put(key,ce));
    }

    public void expireCache () {
        super.clear();
    }

    public Object expireElement (Object o) {
        return (super.remove(o));   
    }
    


}