
public class CircularByteBuffer {
    public byte buffer[];
    public int bufferSize = 256;
    public int currReadPointer = 0;
    public int currWritePointer = 0;
    public int currCopyPointer = 0;

    public CircularByteBuffer(int bufSize)
    {
        bufferSize = bufSize;
        buffer = new byte[bufferSize];
    }

    public CircularByteBuffer()
    {
        buffer = new byte[bufferSize];
    }

    public int bytesToRead()
    {
        if (currWritePointer>=currReadPointer)
            return (currWritePointer - currReadPointer);
        else
            return (bufferSize - currReadPointer + currWritePointer);
    }

    public int bytesToCopy()
    {
        if (currWritePointer>=currCopyPointer)
            return (currWritePointer - currCopyPointer);
        else
            return (bufferSize - currCopyPointer + currWritePointer);
    }

    public void copyReset()
    {
        currCopyPointer = currReadPointer;
    }

    public void write(byte b) throws CircularBufferFullException
    {
        if (bytesToRead()==bufferSize-1) throw new CircularBufferFullException("Circular Buffer Size Exceeded!");
        buffer[currWritePointer] = b;
        currWritePointer++;
        if (currWritePointer==bufferSize) currWritePointer=0;
    }

    public int read()
    {
        if (bytesToRead()==0) return (-1);
        byte b = buffer[currReadPointer];
        if (currReadPointer==currCopyPointer) {
            currReadPointer++;
            currCopyPointer++;
            if (currReadPointer==bufferSize) currReadPointer=0;
            if (currCopyPointer==bufferSize) currCopyPointer=0;
        }
        else {
            currReadPointer++;
            if (currReadPointer==bufferSize) currReadPointer=0;
        }
        return (int)b;
    }

    public int copy()
    {
        if (bytesToCopy()==0) return (-1);
        byte b = buffer[currCopyPointer];
        currCopyPointer++;
        if (currCopyPointer==bufferSize) currCopyPointer=0;
        return (int)b;
    }

}
