import java.util.*;
import java.io.*;
import com.ms.xml.om.*;
import com.ms.xml.util.*;
import com.ms.xml.parser.*;
import java.lang.reflect.*;
import java.sql.*;


public abstract class TDLTag {
    private Hashtable localTagCollection;
    protected Template tdlTemplate;
    private Element tdlElement;
    private Hashtable tdlLocalProperties;
    private String clientIP;
    private Connection conn;
    
    public TDLTag() {
    }
    

//TAG API    


    public int unscrambleUser (int extusid) {
        int res = 1;
        StringTokenizer st = new StringTokenizer (clientIP,".");
        String tok = st.nextToken();
        while (st.hasMoreTokens()) {
            tok = st.nextToken();
            res *= Integer.parseInt(tok);
        }
 //       return extusid - res;
        return extusid;
    }

    public InputStream getFileHandle (String fileName) {
        return (tdlTemplate.getFileStream (fileName));   
    } 

    public void removeAttrValue (String name) {
        tdlLocalProperties.remove(name);
    }

    public String getStringAttrValue(String name) {
        String str = (String)tdlLocalProperties.get(name);
        if (str==null) return "";
        else return str;
    }
    
    public int getIntAttrValue(String name) {
        Integer integer;
        String strValue = (String)tdlLocalProperties.get(name);
        if (strValue==null) return 0;
        
        try {
            integer = new Integer(strValue);
        } 
        catch (NumberFormatException e) {
            integer = new Integer(0);
        }
        return (integer.intValue());
    }
    
    public float getFloatAttrValue(String name) {
        Float f;
        String strValue = (String)tdlLocalProperties.get(name);
        if (strValue==null) return 0;

        try {
            f = new Float(strValue);
        } 
        catch (NumberFormatException e) {
            f = new Float(0);
        }
        return (f.floatValue());
    }
    
    public boolean setStringAttrValue(String name, String value) {
        if (tdlLocalProperties.put(name,value)==null) return false; else return true;
    }

    public boolean setIntAttrValue(String name, int value) {
        String str = Integer.toString(value);
        return setStringAttrValue(name,str);
    }

    public boolean setFloatAttrValue(String name, float value) {
        String str = Float.toString(value);
        return setStringAttrValue(name,str);
    }


// to do boooolleann 


    
    
    final String parseChildren() {
        return parseTagTree(tdlTemplate, tdlElement, tdlLocalProperties);
    }

//end of TAg API 
    public String parseTagTree(Template template, Element element, Hashtable localProperties) {
        String res = "";  
        Enumeration e = element.getElements();
        Element el;
        while (e.hasMoreElements()) {
            el = (Element)e.nextElement();

            if (el.getType()==Element.PCDATA) {
                res += getTagPCDATA(el); 
            }
            else {
                res += handleTagWrapper(el, template, localProperties);
            }
        }
        return res;
    }


    public void setConn (Connection cn) {
        conn = cn;
    }
    public Connection getConn () {
        return conn;
    }


    public void setClientIP (String ip) {
        clientIP = ip;
    }

    public void setTagCollection(Hashtable tags) {
        localTagCollection = tags;
    }

    public void setTDLTemplate(Template t) {
        tdlTemplate = t;
    }

    public void setTDLElement(Element e) {
        tdlElement = e;
    }
    
    public void setTDLLocalProperties(Hashtable lp) {
        tdlLocalProperties = lp;
    }


    private String handleTagWrapper(Element element, Template template, Hashtable localProperties) {

        Hashtable attributes = new Hashtable();
        Enumeration attrEnum = element.getAttributes();
        int i;
        String problem="";

        if (element.getTagName()==null) return "";
        else problem = element.getTagName().getName();

        if (!localTagCollection.containsKey(element.getTagName().getName())) {
            return "";
        }

        while (attrEnum.hasMoreElements())
        {
            Attribute attr = (Attribute)attrEnum.nextElement();
            attributes.put(attr.getName().getName(), attr.getValue());
        }

        MethodInvocation mI = (MethodInvocation)localTagCollection.get(element.getTagName().getName());
        TDLTag tdlTag;
        try {
            if (mI.isPersistent) {
                tdlTag = template.getPersistentObject (mI.className);
                tdlTag.setTDLElement(element);
                tdlTag.setTDLLocalProperties(localProperties);
                tdlTag.setClientIP(clientIP);
                tdlTag.setConn (conn);
            }
            else {
                Class cl = Class.forName(mI.className);
                tdlTag = (TDLTag)cl.newInstance();
                tdlTag.setTagCollection(localTagCollection);

                tdlTag.setTDLTemplate(template);
                tdlTag.setTDLElement(element);
                tdlTag.setTDLLocalProperties(localProperties);
                tdlTag.setClientIP(clientIP);
                tdlTag.setConn (conn);
            }
    		Class[] params = new Class[mI.numberOfArguments];

       		for (i = 0; i < mI.numberOfArguments; i++) params[i] = (new String()).getClass();
    		Method tagmethod = tdlTag.getClass().getMethod(mI.methodName, params);

            Object argumentTable[] = new Object[mI.numberOfArguments];

            for (i=0; i<mI.numberOfArguments; i++) argumentTable[i]=(String)attributes.get(mI.argumentList[i]);


            String s = (String)tagmethod.invoke(tdlTag,argumentTable);
    		return(s);

		} catch (Exception e) {
		    ExceptionHandler.handleException(e,"Tag Process exception on "+problem+" : "+e.getMessage());
		    return ("");
		}
    }
    
    
    private String getTagPCDATA(Element element) {
        return element.getText();
    }
    
 
   public String evaluateExpression (String in) {
        try {
            Float f = new Float(in);
            return (in);
        } catch (NumberFormatException e) {
            return evaluateComplexExpression(in);
        }
    }

//evaluator
    static final int functionUNKNOWN=0;

    static final int functionEQUAL=101;
    static final int functionDIFFERENT=102;
    static final int functionGREATER=103;
    static final int functionLESS=104;
    static final int functionNOGREATER=105;
    static final int functionNOLESS=106;
    
    static final int functionOR=201;
    static final int functionAND=202;
    static final int functionNOT=203;
    
    static final int functionTOLOWER=301;
    static final int functionTOUPPER=302;
    static final int functionTRIM=303;
    static final int functionNOSPACE=304;
    
    static final int functionADD=401;
    static final int functionSUM=402;
    static final int functionMULTIPLY=403;
    static final int functionDIVIDE=404;
    static final int functionSUBSTRACT=405;

    static final int functionMINIMUM=501;
    static final int functionMAXIMUM=502;
    
    static final char variablePrefix='$';
    static final char callPrefix='`';
    static final char functionPrefix='%';
    static final char variableLeftBracket='{';
    static final char variableRightBracket='}';
    static final char parameterLeftBracket='(';
    static final char parameterRightBracket=')';
    static final char parameterSeparator=',';
 
    private String evaluateComplexExpression(String expression)
    {
        String result;
        String subexpression;
        boolean endOfExpression;
        boolean booleanResult;
        boolean booleanFunctionType;
        boolean numericFunctionType;
        double numericResult;
        double numericArgument;
        char ch;
        int position;
        int positionBegin;
        int positionEnd;
        int bracketLevel;
        int i;
        String functionName;
        int functionType;
        String argument1;
        String argument2;
        // Expression evaluation
        // BNF-like grammar:
        //  Expression :=   Variable |
        //                  Constant |
        //                  %equ[al](ExpressionList) |
        //                  %dif[ferent] (ExpressionList) |
        //                  %or (ExpressionList) |
        //                  %and (ExpressionList) |
        //                  %gre[ater] (ExpressionList) |
        //                  %les[s] (ExpressionList) |
        //                  %nog[reater] (ExpressionList) |
        //                  %nol[ess] (ExpressionList) |
        //                  %tol[ower] (ExpressionList) |
        //                  %tou[pper] (ExpressionList) |
        //                  %tri[m] (ExpressionList) |
        //                  %nos[pace] (ExpressionList) |
        //                  %add(ExpressionList) |
        //                  %mul[tiply](ExpressionList) |
        //                  %div[ide](ExpressionList) |
        //                  %sub[stract](ExpressionList) |
        //                  %sum(ExpressionList) |
        //                  %min[imum](ExpressionList) |
        //                  %max[imum](ExpressionList) |
        //                  %not (Expression)
        //
        // ExpressionList ::= Expression ; ExpressionList |
        //                    Expression
        //
        // Variable ::= ${variable_name}    |
        //              `class.method(par)` |
        //              $$      |
        //              ${$}    |
        //              $`      |
        //              ${`}    |
        //              $%      |
        //              ${%}    |
        //              $;      |
        //              ${;}    |
        //              $}      |
        //              ${quot} |
        //              ${true} |
        //              ${false}
        //
        // Constant ::= string  |
        //              %{string}
        //
        // Boolean true is equivalent to any non-empty string, while false - to empty string
        
        result="";
        endOfExpression=false;
        position=0;
        // main loop -  getting one character; if any prefix, going down to
        //              end-prefix and calling a sub-procedure; after the call,
        //              adding sub-result to result
        //              if not prefix, adding a character to result and going down
        
        if (expression.length()==0) return("");
        while (!endOfExpression)
        {
            ch=getOneChar(expression,position,' ');
            if (isPrefix(ch))
            {
                switch (ch)
                {
                case variablePrefix:
                    // getting special character or variable
                    position++;
                    if (position>=expression.length()) 
                    {
                        result+=variablePrefix;
                        break;
                    }
                    ch=getOneChar(expression,position,' ');
                    if (isPrefix(ch))
                        result+=ch;
                    else if (ch==parameterSeparator)
                        result+=ch;
                    else if (ch==variableRightBracket)
                        result+=ch;
                    else if (ch==variableLeftBracket)
                    {
                        // getting a variable or special key
                        subexpression="";
                        position++;
                        ch=getOneChar(expression,position,variableRightBracket);
                        while ( (position<expression.length()) &&
                                (ch!=variableRightBracket) )
                        {
                            subexpression+=ch;
                            position++;
                            ch=getOneChar(expression,position,variableRightBracket);
                        }
                        // evaluating variable name
                        if (subexpression.equals(""+variablePrefix))
                            result+=variablePrefix;
                        else if (subexpression.equals(""+callPrefix))
                            result+=callPrefix;
                        else if (subexpression.equals(""+functionPrefix))
                            result+=functionPrefix;
                        else if (subexpression.equals(""+parameterSeparator))
                            result+=parameterSeparator;
                        else if (subexpression.equalsIgnoreCase("quot"))
                            result+="\"";
                        else if (subexpression.equalsIgnoreCase("true"))
                            result+="TRUE";
                        else if (subexpression.equalsIgnoreCase("false"))
                            result+="";
                        else 
                            result+=evaluateVariable(subexpression);
                    }
                    else 
                    {
                        // $ character is not a recognized prefix
                        // treating it as normal character
                        position--;
                        result+=variablePrefix;
                    }
                    break;
                case functionPrefix:
                    // invoking a function
                    functionName="";
                    // getting function name up to first non-letter character
                    position++;
                    ch=getOneChar(expression, position, parameterLeftBracket);
                    if (ch==variableLeftBracket)
                    {
                        // constant string up to "}" all prefixes are normal characters
                        position++;
                        ch=getOneChar(expression, position, variableRightBracket);
                        while (ch!=variableRightBracket)                  
                        {
                            result+=ch;
                            position++;
                            ch=getOneChar(expression, position, variableRightBracket);
                        }    
                    }
                    else
                    {
                        while ( (position<expression.length()) &&
                                (isLetterOrSpace(ch)) )
                        {
                            if (ch!=' ') functionName+=ch;
                            position++;
                            ch=getOneChar(expression, position, parameterLeftBracket);
                        }
                        // omitting all spaces between name and arguments
                        while ( (position<expression.length()) &&
                                (ch==' ') )
                        {
                            position++;
                            ch=getOneChar(expression, position, parameterLeftBracket);
                        }
                        // checking function type
                        booleanFunctionType=false;
                        numericFunctionType=false;
                        if      (functionName.toLowerCase().startsWith("not"))
                        {
                            functionType=functionNOT;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }    
                        else if (functionName.toLowerCase().startsWith("equ"))
                        {
                            functionType=functionEQUAL;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("dif"))
                        {
                            functionType=functionDIFFERENT;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("or"))
                        {
                            functionType=functionOR;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("and"))
                        {
                            functionType=functionAND;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("tol"))
                        {
                            functionType=functionTOLOWER;
                            booleanFunctionType=false;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("tou"))
                        {
                            functionType=functionTOUPPER;
                            booleanFunctionType=false;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("tri"))
                        {
                            functionType=functionTRIM;
                            booleanFunctionType=false;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("nos"))
                        {
                            functionType=functionNOSPACE;
                            booleanFunctionType=false;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("gre"))
                        {
                            functionType=functionGREATER;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("les"))
                        {
                            functionType=functionLESS;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("nog"))
                        {
                            functionType=functionNOGREATER;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("nol"))
                        {
                            functionType=functionNOLESS;
                            booleanFunctionType=true;
                            numericFunctionType=false;
                        }
                        else if (functionName.toLowerCase().startsWith("add"))
                        {
                            functionType=functionADD;
                            booleanFunctionType=false;
                            numericFunctionType=true;
                        }
                        else if (functionName.toLowerCase().startsWith("sum"))
                        {
                            functionType=functionSUM;
                            booleanFunctionType=false;
                            numericFunctionType=true;
                        }
                        else if (functionName.toLowerCase().startsWith("mul"))
                        {
                            functionType=functionMULTIPLY;
                            booleanFunctionType=false;
                            numericFunctionType=true;
                        }
                        else if (functionName.toLowerCase().startsWith("div"))
                        {
                            functionType=functionDIVIDE;
                            booleanFunctionType=false;
                            numericFunctionType=true;
                        }
                        else if (functionName.toLowerCase().startsWith("sub"))
                        {
                            functionType=functionSUBSTRACT;
                            booleanFunctionType=false;
                            numericFunctionType=true;
                        }
                        else if (functionName.toLowerCase().startsWith("max"))
                        {
                            functionType=functionMAXIMUM;
                            booleanFunctionType=false;
                            numericFunctionType=true;
                        }
                        else if (functionName.toLowerCase().startsWith("min"))
                        {
                            functionType=functionMINIMUM;
                            booleanFunctionType=false;
                            numericFunctionType=true;
                        }
                        else
                            functionType=functionUNKNOWN;
                        // getting a begin of argument list - "(" character
                        if (ch!=parameterLeftBracket)
                        {
                            // no arguments - error
                            functionType=functionUNKNOWN;
                        }
                        if (functionType==functionUNKNOWN)
                        {
                            // sending just read name directly to output - bad function name
                            result+=functionPrefix+functionName+ch;
                        }
                        else if (functionType==functionNOT)
                        {
                            // NOT function operates on one argument
                            // getting the argument up to last ')'
                            argument1="";
                            position++;
                            ch=getOneChar(expression,position,parameterRightBracket);
                            bracketLevel=1;
                            while ( (position<expression.length()) && 
                                    (bracketLevel>0))
                                {
                                    if (ch==parameterRightBracket) 
                                    {
                                        bracketLevel--;
                                        if (bracketLevel>0)
                                            argument1+=ch;
                                    }        
                                    else if (ch==parameterLeftBracket) 
                                    {
                                        bracketLevel++;
                                        argument1+=ch;
                                    }    
                                    else
                                    {
                                        argument1+=ch;
                                    }
                                    position++;
                                    ch=getOneChar(expression,position,parameterRightBracket);
                                }
                            // argument read, evaluating
                            argument2=evaluateExpression(argument1);
                            // non-empty value to empty
                            // empty value to "TRUE"
                            if (argument2.length()==0) result+="TRUE";
                            // end NOT evaluation
                        }
                        else
                        {
                            // other functions are exhausting - N arguments
                            // getting first argument
                            ch=getOneChar(expression,position,parameterLeftBracket);
                            while (ch!=parameterLeftBracket)
                            {
                                ch=getOneChar(expression,position,parameterLeftBracket);
                                position++;
                                // getting to the beginning of the argument
                            }    
                            bracketLevel=0;    
                            argument1="";
                            booleanResult=false;
                            numericResult=0;
                            if (functionType==functionMULTIPLY)
                                numericResult=1;
                            ch=getOneChar(expression,position,parameterRightBracket);
                            if ( (ch==parameterSeparator) || (ch==parameterLeftBracket) ) position++;
                            positionBegin=position;
                            positionEnd=getArgumentEnd(expression, positionBegin, parameterSeparator, parameterRightBracket);
                            while (positionEnd>positionBegin)
                            {
                                // next argument to evaluate
                                argument2="";
                                while (position<positionEnd)
                                    argument2+=getOneChar(expression,position++,parameterRightBracket);
                                // getting next argument(s) separated by ";" in a loop
                                // up to ")"
                                argument2=evaluateExpression(argument2);
                                if (!booleanFunctionType)
                                    switch (functionType)
                                    {
                                        case functionTOLOWER:
                                            result+=argument2.toLowerCase();
                                            break;
                                        case functionTOUPPER:
                                            result+=argument2.toUpperCase();
                                            break;
                                        case functionTRIM:
                                            // removes spaces and tabs from both end of the string    
                                            result+=argument2.trim();
                                            break;
                                        case functionNOSPACE:
                                            // removes spaces and tabs from both end of the string    
                                            for (i=0;i<argument2.length();i++)
                                                if (argument2.charAt(i)!=' ')
                                                    result+=argument2.charAt(i);
                                            break;
                                        case functionADD:
                                        case functionSUM:
                                            // adds N arguments into one argument
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            numericResult+=numericArgument;
                                            break;
                                        case functionMULTIPLY:
                                            // multiplays N arguments into one argument
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            numericResult*=numericArgument;
                                            break;
                                        case functionDIVIDE:
                                            // divides N arguments into one argument
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (bracketLevel==0)
                                                numericResult=numericArgument;
                                            else
                                            {
                                                if (numericArgument==0)
                                                    numericResult=0;
                                                else
                                                    numericResult/=numericArgument;
                                            }        
                                            break;
                                        case functionSUBSTRACT:
                                            // divides N arguments into one argument
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (bracketLevel==0)
                                                numericResult=numericArgument;
                                            else
                                                numericResult-=numericArgument;
                                            break;
                                        case functionMAXIMUM:
                                            // getting the highest argument
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (bracketLevel==0)
                                                numericResult=numericArgument;
                                            else
                                            {
                                                if (numericResult<numericArgument)
                                                    numericResult=numericArgument;
                                            }        
                                            break;
                                        case functionMINIMUM:
                                            // getting the lowest argument
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (bracketLevel==0)
                                                numericResult=numericArgument;
                                            else
                                            {
                                                if (numericResult>numericArgument)
                                                    numericResult=numericArgument;
                                            }        
                                            break;
                                        default:
                                            break;
                                    }    
                                if (bracketLevel>0)
                                // boolean-resulting function if at least two arguments
                                {
                                    // performing a function on two successing arguments
                                    switch (functionType)
                                    {
                                        case functionEQUAL:
                                            if (!argument1.equals(argument2))
                                                booleanResult=false;
                                            break;
                                        case functionDIFFERENT:
                                            if (argument1.equals(argument2))
                                                booleanResult=false;
                                            break;
                                        case functionOR:
                                            if (argument2.length()>0)
                                                booleanResult=true;
                                            break;
                                        case functionAND:
                                            if (argument2.length()==0)
                                                booleanResult=false;
                                            break;
                                        case functionGREATER:
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (numericResult>numericArgument )
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            break;    
                                        case functionLESS:
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (numericResult<numericArgument)
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            break;    
                                        case functionNOLESS:
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (numericResult>=numericArgument)
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            break;    
                                        case functionNOGREATER:
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (numericResult<=numericArgument)
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            break;    
                                        default:
                                            break;
                                    }
                                }
                                else
                                // function for first argument
                                {
                                    // performing a function on a first argument
                                    switch (functionType)
                                    {
                                        case functionEQUAL:
                                            booleanResult=true;
                                            break;
                                        case functionDIFFERENT:
                                            booleanResult=true;
                                            break;
                                        case functionOR:
                                            if (argument2.length()>0)
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            break;
                                        case functionAND:
                                            if (argument2.length()==0)
                                                booleanResult=false;
                                            else
                                                booleanResult=true;
                                            break;
                                        case functionGREATER:
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (numericArgument>0)
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            numericResult=numericArgument;    
                                            break;    
                                        case functionLESS:
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (numericArgument<0)
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            numericResult=numericArgument;    
                                            break;    
                                        case functionNOLESS:
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (numericArgument>=0)
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            numericResult=numericArgument;    
                                            break;    
                                        case functionNOGREATER:
                                            try
                                            {
                                                numericArgument=Double.valueOf(argument2).doubleValue();
                                            } catch (NumberFormatException e) {numericArgument=0;} 
                                            if (numericArgument<=0)
                                                booleanResult=true;
                                            else
                                                booleanResult=false;
                                            numericResult=numericArgument;    
                                            break;    
                                        default:
                                            break;
                                    }
                                }
                                bracketLevel++;    
                                argument1=argument2;    
                                // looping for next argument or end of call
                                ch=getOneChar(expression,position,parameterRightBracket);
                                if ( ch==parameterSeparator) position++;
                                positionBegin=position;
                                positionEnd=getArgumentEnd(expression, positionBegin, parameterSeparator, parameterRightBracket);
                            }
                            if (booleanResult && booleanFunctionType)
                                result+="true"; 
                            if (numericFunctionType)
                            {
                                i=(new Double(numericResult)).intValue();
                                if (i==numericResult)
                                    // if result is an integer, integer output
                                    result+=(new Integer(i)).toString();
                                else
                                    // else real output
                                    result+=(new Double(numericResult)).toString();
                            }    
                        }
                    }       
                    break;
                case callPrefix:
                    // invoking Java call
                    ch=getOneChar(expression,position,variablePrefix);
                    if ( ch==callPrefix) position++;
                    positionBegin=position;
                    positionEnd=getArgumentEnd(expression, positionBegin, callPrefix, callPrefix);
                    if (positionEnd>positionBegin)
                    {
                        // an argument to evaluate
                        argument1="";
                        while (position<positionEnd)
                            argument1+=getOneChar(expression,position++,parameterRightBracket);
                        ch=getOneChar(expression,position,parameterRightBracket);
                        while ( ch!=callPrefix) 
                        {
                            position++;
                            ch=getOneChar(expression,position,callPrefix);
                        }    
                        argument2=evaluateExpression(argument1);
                        // resulting string is a Java call
                        result+=callJavaMethod(argument2);
                    }    
                    break;
                default:
                    break;
                }    
            }
            else
                result+=ch;
            position++;
            if (position>=expression.length()) endOfExpression=true;
        }
        
        return (result);
    }

    private String callJavaMethod (String arg) {
        
        MethodInvocation mI = new MethodInvocation (arg);
        try {
            Class cl = Class.forName(mI.className);        
            Object ob = cl.newInstance();
       		Class[] params = new Class[mI.numberOfArguments];
            Object argumentTable[] = new Object[mI.numberOfArguments];
            int i;
            Float f;
   	    	for (i = 0; i < mI.numberOfArguments; i++) {
                try {
                     f = new Float(mI.argumentList[i]); 
                     params[i] = Float.TYPE;
                     argumentTable[i] = f;
                } catch (NumberFormatException e) {
                     params[i] = (new String()).getClass();
                     argumentTable[i]=mI.argumentList[i];
                }
            }
   		    Method me = ob.getClass().getMethod(mI.methodName, params);
       		Class resc = me.getReturnType();

   	    	if (resc.equals ((new String()).getClass()) ) {
                return ((String)me.invoke(cl,argumentTable));
            }
            else {
                f = (Float)me.invoke(cl,argumentTable);
                return (f.toString());
            } 
        } catch (Exception e) {
            ExceptionHandler.handleException(e,"Evaluate.callJavaMethod");
            return "";
        }
        

    }
    

    private String evaluateVariable(String variableName)
    {
        return ((String)tdlLocalProperties.get(variableName));
    }  
    
    private char getOneChar(String expression, int position, char ch1)
    {   
        char ch;
        if (position<expression.length())
            ch=expression.charAt(position);
        else
            ch=ch1;
        return (ch);
    }
    
    private int getArgumentEnd(String expression, int positionBegin, char suffix1, char suffix2)
    {
        int position;
        char ch;
        position=positionBegin;
        ch=getOneChar(expression,position,parameterRightBracket);
        while ( (ch!=suffix1) && (ch!=suffix2) )
        {
            if (isPrefix(ch))
            {
                if (ch==variablePrefix)
                {
                    // $ prefixes one character or {string}
                    position++;
                    ch=getOneChar(expression, position, suffix1);
                    if (ch==variableLeftBracket)
                    {
                        position=1+getArgumentEnd(expression,position+1, variableRightBracket, variableRightBracket);
                        ch=getOneChar(expression, position, suffix1);
                    }    
                }   
                // ` prefixes Java call up to `
                if (ch==callPrefix)
                {
                    // ` prefixes Java call up to '`'
                    position=1+getArgumentEnd(expression,position+1, callPrefix, callPrefix);
                    ch=getOneChar(expression, position, suffix1);
                }   
                if (ch==functionPrefix)
                {
                    // % prefixes function call to ')'
                    position=1+getArgumentEnd(expression,position+1, parameterRightBracket, parameterRightBracket);
                    ch=getOneChar(expression, position, suffix1);
                }   
                
            }
            else
                position++;
            ch=getOneChar(expression, position, suffix1);

        }    
        return position;
    }
    
    private boolean isPrefix(char prefix)
    {
        if (prefix == variablePrefix) return (true);
        if (prefix == callPrefix) return (true);
        if (prefix == functionPrefix) return (true);
        return (false);        
    }

    private boolean isLetter(char ch)
    {
        if ((ch>='a') && (ch<='z')) return (true);
        if ((ch>='A') && (ch<='Z')) return (true);
        return (false);
    }
    
    private boolean isLetterOrSpace(char ch)
    {
        if (isLetter(ch)) return (true);
        if (ch==' ') return (true);
        return (false);
    }




}