import java.util.*;
import java.net.*;
import java.io.*;
import java.sql.*;

public class Template {

    private boolean isPropLoadedFlag = false;

    private Hashtable templateProperties = new Hashtable();
    private Hashtable persistentObjectCollection = new Hashtable();
    private Cache loadedFiles = new Cache();
//    private Connection con = null;    
    
    public Template (AcceptedTagList tc) {
        createPersistentObjects(tc);
    }

/*    public Connection getCon () {
        return con;
    }
  */  
/*    public void setCon (Connection lCon) {
        con = lCon;
    }
*/
/*    public void closeCon () {
        try {
            if (con != null) con.close();
        } catch (SQLException e) {
            ExceptionHandler.handleException(e, "closeCon");
        }
    }
*/

	public TDLTag getPersistentObject(String className) {
	    return ((TDLTag)persistentObjectCollection.get(className));
	}

	private void createPersistentObjects (AcceptedTagList tagCollection) {

	      for (Enumeration e = tagCollection.elements();e.hasMoreElements();) {
	         MethodInvocation mI = (MethodInvocation)e.nextElement();
	         if (mI.isPersistent) {
                try {
    	            Class cl = Class.forName(mI.className);
	                TDLTag xml = (TDLTag)cl.newInstance();
	                xml.setTagCollection(tagCollection);
	                xml.setTDLTemplate(this);
	                persistentObjectCollection.put(mI.className,xml);
	            } catch (Exception ex) {
	                   ExceptionHandler.handleException(ex, "create pers objects");
	            }
	         }
	      }
	}

    public int getTemplateCacheSize () {
        Enumeration en = loadedFiles.keys();
        String tmpS;
        int cs = 0;
        while (en.hasMoreElements()) {
            tmpS = (String)loadedFiles.get(en.nextElement());
            if (tmpS!=null) cs+=tmpS.length();
        }
        return cs;
/*        return (loadedFiles.getCacheSize());*/
    }

    public int getNoOfKeys () {
        return (loadedFiles.size());
    }

    public boolean isTemplateCacheActive () {
        return loadedFiles.isCacheActive();
    }

    public void expireFileFromCache (String fileName) {
        loadedFiles.remove(fileName);
    }

    public void expireTemplateCache () {
        loadedFiles.expireCache();    
    }

    public void setTCacheInactive () {
        loadedFiles.setCacheInactive();
    }
    
    public void setTCacheActive () {
        loadedFiles.setCacheActive();
    }

    public void setTMaxHits (int mh) {
        loadedFiles.setMaxHits(mh);
    }

    public void setTMaxTime (long mt) {
        loadedFiles.setMaxTime(mt);
    }
    
    public int getTMaxHits () {
        return loadedFiles.getMaxHits();
    }

    public long getTMaxTime () {
        return loadedFiles.getMaxTime();
    }
    
    
    
    public InputStream getTemplateFileStream (String fileName) {
        return (getFileStream(fileName));
    }
    
    
    private boolean isFileCacheable (String contents) {
        int begs = contents.indexOf("<?XML");
        int ends = contents.indexOf("?>",begs);
        if ((begs == -1) || (ends == -1)) return false;
        String xml = contents.substring(begs,ends+1);
        begs = xml.indexOf("CACHE");
        if (begs == -1) return false;
        int ba = xml.indexOf("\"",begs)+1;
        int ea = xml.indexOf("\"",ba);
        if ((ba == -1) || (ea == -1)) return false;
        String answer = xml.substring(ba,ea);
        if (answer.equals("NO")) return false;
        else return true;
    }

    public InputStream getFileStream(String fileName) {
        String fileContents="";
        
        if (loadedFiles.containsKey(fileName)) 
            fileContents = (String)loadedFiles.get(fileName);
        else {
            fileContents = getStringFromUrl (getProperty(Constants.TemplateLocationProperty)+fileName);
            if ((isFileCacheable(fileContents)) && (isTemplateCacheActive()))
                loadedFiles.put(fileName, fileContents);
        }
        return new StringBufferInputStream(fileContents);
    }


    public Enumeration getPropertyKeys() {
        return templateProperties.keys();
    }
    public void addToProperties(Object k, Object v) {
        templateProperties.put(k,v);
    }
    public Object getProperty(Object k) {
        return templateProperties.get(k);
    }

    public boolean isPropLoaded() {
        return isPropLoadedFlag;
    }

    public void setQueryParameters(String qs) {
        StringTokenizer st = new StringTokenizer (qs,Constants.QueryParametersSeparator);
        while (st.hasMoreElements()) {
            StringTokenizer kv = new StringTokenizer(st.nextToken(),Constants.QueryParamValuesSeparator);
            addToProperties(kv.nextToken(),kv.nextToken());
        }
    }

    public void loadProperties() throws PropertyLoadException {
        if (isPropLoaded()) return;
        BufferedReader br;
        try {
            URL url = new URL(getProperty(Constants.TemplateLocationProperty)+Constants.TemplatePropertyFileName);
            br = new BufferedReader(new InputStreamReader(new DataInputStream(url.openConnection().getInputStream())));
            String line, propName, propValue;

            do {
                line = br.readLine();
                if (line==null) { ExceptionHandler.debugInfo("Template properties file not found."); return; }
            }
            while (!line.equals(Constants.PropertyFileDelimiterBeg));

            while ((line = br.readLine())!=null) {
                line = line.trim();
                if (line.equals(Constants.PropertyFileDelimiterEnd)) break;
                propName = line.substring(0, line.indexOf(Constants.PropertyValueSeparator)).trim();
                propValue = line.substring(line.indexOf(Constants.PropertyValueSeparator)+1,line.length()).trim();
                addToProperties(propName,propValue);
            }
            isPropLoadedFlag=true;
        }
        catch (Exception e) {
            throw new PropertyLoadException("Cannot load template properties."+e.getMessage());
        }
    }

    public Hashtable cloneTemplateProperties() {
        Hashtable ht = new Hashtable();
        Enumeration enum = templateProperties.keys();
        Object o;
        while (enum.hasMoreElements()) {
            o = enum.nextElement();
            ht.put(o, templateProperties.get(o));
        }
        return ht;
    }


	private String getStringFromUrl (String urlAddress) {
        
        try {
    	    URL url = new URL(urlAddress);

        // Directly use implementation to get 
            HttpImpl conn = new HttpImpl(url);

            // Retrieve contents
            String ct = conn.getContentType();
            String st = (String)conn.getContent();
            return st;

            // Display what we've found
            // System.out.println(obj);
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
	}

}

