
import java.io.*;
import java.util.*;

public class WWInlineInputStream extends FilterInputStream {

    CircularByteBuffer circBuffer = new CircularByteBuffer(1024);
    Template template;
    WWInlineInputStream embeddedStream = null;
    boolean caching;

    public WWInlineInputStream(InputStream is, Template t)
    {
        super(is);
        int b;

        template = t;
        
        try {
            for (int i=0; i<63; i++) {
                try {
                    b=super.read();
                }
                catch (IOException e)
                {
                    b=-1;
                }
                if (b==-1) break;
                circBuffer.write((byte)b);
            }
        }
        catch (CircularBufferFullException e)
        {
            System.out.println("Internal WW Inline Input Stream error.");
            e.printStackTrace();
        }


    }

    public void fillCircBuffer() throws IOException, CircularBufferFullException
    {
        int b = super.read();
        if (b!=-1) circBuffer.write((byte)b);
    }

   
    public boolean shouldBeInlined()
    {
        circBuffer.copyReset();
        int c = circBuffer.bytesToCopy();
        byte b[] = new byte[c];
        int bi;
        for (int i =0 ; i<c ; i++ ) {
            bi = circBuffer.copy();
            if (bi==-1) break;
            b[i]=(byte)bi;
        }
        String s = new String(b);

        int indSpace, indGreater, indQuot1, indQuot2;

        indSpace   = s.indexOf(' ');
        indGreater = s.indexOf('>');
        indQuot1   = s.indexOf('"');
        if (indQuot1!=-1) indQuot2   = s.indexOf('"',indQuot1+1); else indQuot2 = -1;

        if ((indSpace==-1)||(indGreater==-1)||(indQuot1==-1)||(indQuot2==-1)) return false;
        if (indQuot2>=indGreater) return false;

        if (indGreater < 8) return false;
        if (!s.substring(0,7).equals(Constants.InlineTagName)) return false;

        String fileName = s.substring(indQuot1+1, indQuot2);

        if (fileName.equals("")) return false;
            
        return true;
    }

// Can be executed ONLY if should be Inlined returns TRUE. Can crash system otherwise.
    public String getFileNameAndCutTag() {
        byte b[] = new byte[circBuffer.bytesToRead()];
        int bi = 0;
        for (int i =0 ; bi!=(int)'>' ; i++ ) {
            bi = circBuffer.read();
            if (bi==-1) break;
            b[i]=(byte)bi;
        }
        String s = new String(b);

        int indQuot1   = s.indexOf('"');
        int indQuot2   = s.indexOf('"', indQuot1+1);

        if ((indQuot1==-1)||(indQuot2==-1)) return "";

        String fileName = s.substring(indQuot1+1, indQuot2);
        return fileName;
    }


    public int read() throws IOException
    {
        int res = -1;
        
        if (embeddedStream!=null) res = embeddedStream.read();
        if (res != -1) return res;

        embeddedStream = null;
        
        res = circBuffer.read();

        try {
            fillCircBuffer();
        }
        catch (CircularBufferFullException e)
        {
            System.out.println("WWInlineInputStream internal error.");
            e.printStackTrace();
        }
        char c = (char)res;

        if (c=='<')  {
            if (shouldBeInlined()) {
                 embeddedStream = new WWInlineInputStream(template.getFileStream(getFileNameAndCutTag()), template);
                 return (read());
            }
        }

        return (res);
    }


    public int read(byte b[], int off, int len) throws IOException
    {
	if (len <= 0) {
	    return 0;
	}

	int c = read();
	if (c == -1) {
	    return -1;
	}
	b[off] = (byte)c;

	int i = 1;
	try {
	    for (; i < len ; i++) {
		c = read();
		if (c == -1) {
		    break;
		}
		if (b != null) {
		    b[off + i] = (byte)c;
		}
	    }
	} catch (IOException ee) {
	}
	return i;
    }
}
