
import java.io.*;
import java.util.*;

public class XMLEncodeInputStream extends FilterInputStream {

    int nextToReturn = -1;
    CircularByteBuffer circBuffer = new CircularByteBuffer(1024);
    Vector XMLElementVector;

    public XMLEncodeInputStream(InputStream is, Vector XMLElements)
    {
        super(is);
        int b;

        XMLElementVector = XMLElements;

        try {
            for (int i=0; i<63; i++) {
                try {
                    b=super.read();
                }
                catch (IOException e)
                {
                    b=-1;
                }
                if (b==-1) break;
                circBuffer.write((byte)b);
            }
        }
        catch (CircularBufferFullException e)
        {
            System.out.println("Internal XML Input Stream error.");
            e.printStackTrace();
        }


    }

    public void fillCircBuffer() throws IOException, CircularBufferFullException
    {
        int b = super.read();
        if (b!=-1) circBuffer.write((byte)b);
    }

    public boolean shouldNotBeEncoded()
    {
        circBuffer.copyReset();
        int c = circBuffer.bytesToCopy();
        byte b[] = new byte[c];
        int bi;
        for (int i =0 ; i<c ; i++ ) {
            bi = circBuffer.copy();
            if (bi==-1) break;
            b[i]=(byte)bi;
        }
        String s = new String(b);
        int indSpace, indGreater, indSlash, ind;

        indSpace   = s.indexOf(' ');
        indGreater = s.indexOf('>');
        
        if (indSpace==-1) indSpace = Integer.MAX_VALUE;
        if (indGreater==-1) indGreater = Integer.MAX_VALUE;
        ind = Math.min(indSpace, indGreater);

        s = s.substring(0,ind);
        if (s.substring(0,1).equals("/")) s=s.substring(1,s.length());

        if (s.charAt(s.length()-1)=='/') s=s.substring(0,s.length()-1);
        
        boolean result = false;

        for (int i=0; i<XMLElementVector.size(); i++) {
            if (s.equals((String)XMLElementVector.elementAt(i))) { result = true; break; }
        }
        return result;
    }


    public int read() throws IOException
    {
        int res;
        if (nextToReturn!=-1)
        {
            res = nextToReturn;
            nextToReturn = -1;
        }
        else
        {
            res = circBuffer.read();
            try {
                fillCircBuffer();
            }
            catch (CircularBufferFullException e)
            {
                System.out.println("XMLEncodeInputStream internal error.");
                e.printStackTrace();
            }
            char c = (char)res;

            if (c=='<')  {
                if (!shouldNotBeEncoded()) {
                    res=(int)'^';  nextToReturn=(int)'^';
                }
            }
            if (c=='^')  { res=(int)'\\'; nextToReturn=(int)'^';  }
            if (c=='\\') { res=(int)'\\'; nextToReturn=(int)'\\'; }
        }
        return (res);
    }


    public int read(byte b[], int off, int len) throws IOException
    {
	if (len <= 0) {
	    return 0;
	}

	int c = read();
	if (c == -1) {
	    return -1;
	}
	b[off] = (byte)c;

	int i = 1;
	try {
	    for (; i < len ; i++) {
		c = read();
		if (c == -1) {
		    break;
		}
		if (b != null) {
		    b[off + i] = (byte)c;
		}
	    }
	} catch (IOException ee) {
	}
	return i;
    }
}
